using std::cout;  using std::cerr;  using std::endl;  //using std::to_string;  // NEED DELETE, PRE-TEMPLATE C++

#ifndef __CPP__INCLUDED__RPerl__DataStructure__Hash__SubTypes1D_cpp
#define __CPP__INCLUDED__RPerl__DataStructure__Hash__SubTypes1D_cpp 0.007_000

#include <RPerl/DataStructure/Hash/SubTypes1D.h>  // -> ??? (relies on <unordered_map> being included via Inline::CPP's AUTO_INCLUDE config option in RPerl/Inline.pm)

// [[[ TYPE-CHECKING ]]]
// [[[ TYPE-CHECKING ]]]
// [[[ TYPE-CHECKING ]]]

// DEV NOTE: for() loops are statements not expressions, so they can't be embedded in ternary operators, and thus this type-checking must be done with subroutines instead of macros
void integer_hashref_CHECK(SV* possible_integer_hashref)
{
    // DEV NOTE: the following two if() statements are functionally equivalent to the hashref_CHECK() macro, but with integer-specific error codes
    if ( not( SvOK(possible_integer_hashref) ) ) { croak( "\nERROR EIVHVRV00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger_hashref value expected but undefined/null value found,\ncroaking" ); }
    if ( not( SvHROKp(possible_integer_hashref) ) ) { croak( "\nERROR EIVHVRV01, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger_hashref value expected but non-hashref value found,\ncroaking" ); }

    HV* possible_integer_hash;
    integer possible_integer_hash_num_keys;
    integer i;
    HE* possible_integer_hashentry;
    SV* possible_integer_hashentry_value;
    SV* possible_integer_hashentry_key;
    string possible_integer_hashentry_key_string;

    possible_integer_hash = (HV*)SvRV(possible_integer_hashref);
    possible_integer_hash_num_keys = hv_iterinit(possible_integer_hash);

    for (i = 0;  i < possible_integer_hash_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        possible_integer_hashentry = hv_iternext(possible_integer_hash);

        // DEV NOTE: the following if() statement is functionally equivalent to the hashentry_CHECK() macro, but with integer-specific error code
        if (possible_integer_hashentry == NULL) { croak("\nERROR EIVHE00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger_hashentry value expected but undefined/null value found,\ncroaking"); }
        possible_integer_hashentry_value = hv_iterval(possible_integer_hash, possible_integer_hashentry);

        // DEV NOTE: the following two if() statements are functionally equivalent to the integer_CHECK() macro & subroutine, but with hash-specific error codes
        if (not(SvOK(possible_integer_hashentry_value)))
        {
            possible_integer_hashentry_key = hv_iterkeysv(possible_integer_hashentry);
            possible_integer_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_integer_hashentry_key)));  // escape key string for error message
            croak("\nERROR EIVHVRV02, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger value expected but undefined/null value found at key '%s',\ncroaking", possible_integer_hashentry_key_string.c_str());
        }
        if (not(SvIOKp(possible_integer_hashentry_value)))
        {
            possible_integer_hashentry_key = hv_iterkeysv(possible_integer_hashentry);
            possible_integer_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_integer_hashentry_key)));  // escape key string for error message
            croak("\nERROR EIVHVRV03, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger value expected but non-integer value found at key '%s',\ncroaking", possible_integer_hashentry_key_string.c_str());
        }
    }
}

void integer_hashref_CHECKTRACE(SV* possible_integer_hashref, const char* variable_name, const char* subroutine_name)
{
    if ( not( SvOK(possible_integer_hashref) ) ) { croak( "\nERROR EIVHVRV00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger_hashref value expected but undefined/null value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name ); }
    if ( not( SvHROKp(possible_integer_hashref) ) ) { croak( "\nERROR EIVHVRV01, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger_hashref value expected but non-hashref value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name ); }

    HV* possible_integer_hash;
    integer possible_integer_hash_num_keys;
    integer i;
    HE* possible_integer_hashentry;
    SV* possible_integer_hashentry_value;
    SV* possible_integer_hashentry_key;
    string possible_integer_hashentry_key_string;

    possible_integer_hash = (HV*)SvRV(possible_integer_hashref);
    possible_integer_hash_num_keys = hv_iterinit(possible_integer_hash);

    for (i = 0;  i < possible_integer_hash_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        possible_integer_hashentry = hv_iternext(possible_integer_hash);

        if (possible_integer_hashentry == NULL) { croak("\nERROR EIVHE00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger_hashentry value expected but undefined/null value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name); }
        possible_integer_hashentry_value = hv_iterval(possible_integer_hash, possible_integer_hashentry);

        if (not(SvOK(possible_integer_hashentry_value)))
        {
            possible_integer_hashentry_key = hv_iterkeysv(possible_integer_hashentry);
            possible_integer_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_integer_hashentry_key)));  // escape key string for error message
            croak("\nERROR EIVHVRV02, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger value expected but undefined/null value found at key '%s',\nin variable %s from subroutine %s,\ncroaking", possible_integer_hashentry_key_string.c_str(), variable_name, subroutine_name);
        }
        if (not(SvIOKp(possible_integer_hashentry_value)))
        {
            possible_integer_hashentry_key = hv_iterkeysv(possible_integer_hashentry);
            possible_integer_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_integer_hashentry_key)));  // escape key string for error message
            croak("\nERROR EIVHVRV03, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\ninteger value expected but non-integer value found at key '%s',\nin variable %s from subroutine %s,\ncroaking", possible_integer_hashentry_key_string.c_str(), variable_name, subroutine_name);
        }
    }
}

void number_hashref_CHECK(SV* possible_number_hashref)
{
    if ( not( SvOK(possible_number_hashref) ) ) { croak( "\nERROR ENVHVRV00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber_hashref value expected but undefined/null value found,\ncroaking" ); }
    if ( not( SvHROKp(possible_number_hashref) ) ) { croak( "\nERROR ENVHVRV01, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber_hashref value expected but non-hashref value found,\ncroaking" ); }

    HV* possible_number_hash;
    integer possible_number_hash_num_keys;
    integer i;
    HE* possible_number_hashentry;
    SV* possible_number_hashentry_value;
    SV* possible_number_hashentry_key;
    string possible_number_hashentry_key_string;

    possible_number_hash = (HV*)SvRV(possible_number_hashref);
    possible_number_hash_num_keys = hv_iterinit(possible_number_hash);

    for (i = 0;  i < possible_number_hash_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        possible_number_hashentry = hv_iternext(possible_number_hash);

        if (possible_number_hashentry == NULL) { croak("\nERROR ENVHE00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber_hashentry value expected but undefined/null value found,\ncroaking"); }
        possible_number_hashentry_value = hv_iterval(possible_number_hash, possible_number_hashentry);

        if (not(SvOK(possible_number_hashentry_value)))
        {
            possible_number_hashentry_key = hv_iterkeysv(possible_number_hashentry);
            possible_number_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_number_hashentry_key)));  // escape key string for error message
            croak("\nERROR ENVHVRV02, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber value expected but undefined/null value found at key '%s',\ncroaking", possible_number_hashentry_key_string.c_str());
        }
        if (not(SvNOKp(possible_number_hashentry_value) || SvIOKp(possible_number_hashentry_value)))
        {
            possible_number_hashentry_key = hv_iterkeysv(possible_number_hashentry);
            possible_number_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_number_hashentry_key)));  // escape key string for error message
            croak("\nERROR ENVHVRV03, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber value expected but non-number value found at key '%s',\ncroaking", possible_number_hashentry_key_string.c_str());
        }
    }
}

void number_hashref_CHECKTRACE(SV* possible_number_hashref, const char* variable_name, const char* subroutine_name)
{
    if ( not( SvOK(possible_number_hashref) ) ) { croak( "\nERROR ENVHVRV00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber_hashref value expected but undefined/null value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name ); }
    if ( not( SvHROKp(possible_number_hashref) ) ) { croak( "\nERROR ENVHVRV01, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber_hashref value expected but non-hashref value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name ); }

    HV* possible_number_hash;
    integer possible_number_hash_num_keys;
    integer i;
    HE* possible_number_hashentry;
    SV* possible_number_hashentry_value;
    SV* possible_number_hashentry_key;
    string possible_number_hashentry_key_string;

    possible_number_hash = (HV*)SvRV(possible_number_hashref);
    possible_number_hash_num_keys = hv_iterinit(possible_number_hash);

    for (i = 0;  i < possible_number_hash_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        possible_number_hashentry = hv_iternext(possible_number_hash);

        if (possible_number_hashentry == NULL) { croak("\nERROR ENVHE00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber_hashentry value expected but undefined/null value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name); }
        possible_number_hashentry_value = hv_iterval(possible_number_hash, possible_number_hashentry);

        if (not(SvOK(possible_number_hashentry_value)))
        {
            possible_number_hashentry_key = hv_iterkeysv(possible_number_hashentry);
            possible_number_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_number_hashentry_key)));  // escape key string for error message
            croak("\nERROR ENVHVRV02, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber value expected but undefined/null value found at key '%s',\nin variable %s from subroutine %s,\ncroaking", possible_number_hashentry_key_string.c_str(), variable_name, subroutine_name);
        }
        if (not(SvNOKp(possible_number_hashentry_value) || SvIOKp(possible_number_hashentry_value)))
        {
            possible_number_hashentry_key = hv_iterkeysv(possible_number_hashentry);
            possible_number_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_number_hashentry_key)));  // escape key string for error message
            croak("\nERROR ENVHVRV03, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nnumber value expected but non-number value found at key '%s',\nin variable %s from subroutine %s,\ncroaking", possible_number_hashentry_key_string.c_str(), variable_name, subroutine_name);
        }
    }
}

void string_hashref_CHECK(SV* possible_string_hashref)
{
    if ( not( SvOK(possible_string_hashref) ) ) { croak( "\nERROR EPVHVRV00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring_hashref value expected but undefined/null value found,\ncroaking" ); }
    if ( not( SvHROKp(possible_string_hashref) ) ) { croak( "\nERROR EPVHVRV01, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring_hashref value expected but non-hashref value found,\ncroaking" ); }

    HV* possible_string_hash;
    integer possible_string_hash_num_keys;
    integer i;
    HE* possible_string_hashentry;
    SV* possible_string_hashentry_value;
    SV* possible_string_hashentry_key;
    string possible_string_hashentry_key_string;

    possible_string_hash = (HV*)SvRV(possible_string_hashref);
    possible_string_hash_num_keys = hv_iterinit(possible_string_hash);

    for (i = 0;  i < possible_string_hash_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        possible_string_hashentry = hv_iternext(possible_string_hash);

        if (possible_string_hashentry == NULL) { croak("\nERROR EPVHE00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring_hashentry value expected but undefined/null value found,\ncroaking"); }
        possible_string_hashentry_value = hv_iterval(possible_string_hash, possible_string_hashentry);

        if (not(SvOK(possible_string_hashentry_value)))
        {
            possible_string_hashentry_key = hv_iterkeysv(possible_string_hashentry);
            possible_string_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_string_hashentry_key)));  // escape key string for error message
            croak("\nERROR EPVHVRV02, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring value expected but undefined/null value found at key '%s',\ncroaking", possible_string_hashentry_key_string.c_str());
        }
        if (not(SvPOKp(possible_string_hashentry_value)))
        {
            possible_string_hashentry_key = hv_iterkeysv(possible_string_hashentry);
            possible_string_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_string_hashentry_key)));  // escape key string for error message
            croak("\nERROR EPVHVRV03, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring value expected but non-string value found at key '%s',\ncroaking", possible_string_hashentry_key_string.c_str());
        }
    }
}

void string_hashref_CHECKTRACE(SV* possible_string_hashref, const char* variable_name, const char* subroutine_name)
{
    if ( not( SvOK(possible_string_hashref) ) ) { croak( "\nERROR EPVHVRV00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring_hashref value expected but undefined/null value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name ); }
    if ( not( SvHROKp(possible_string_hashref) ) ) { croak( "\nERROR EPVHVRV01, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring_hashref value expected but non-hashref value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name ); }

    HV* possible_string_hash;
    integer possible_string_hash_num_keys;
    integer i;
    HE* possible_string_hashentry;
    SV* possible_string_hashentry_value;
    SV* possible_string_hashentry_key;
    string possible_string_hashentry_key_string;

    possible_string_hash = (HV*)SvRV(possible_string_hashref);
    possible_string_hash_num_keys = hv_iterinit(possible_string_hash);

    for (i = 0;  i < possible_string_hash_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        possible_string_hashentry = hv_iternext(possible_string_hash);

        if (possible_string_hashentry == NULL) { croak("\nERROR EPVHE00, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring_hashentry value expected but undefined/null value found,\nin variable %s from subroutine %s,\ncroaking", variable_name, subroutine_name); }
        possible_string_hashentry_value = hv_iterval(possible_string_hash, possible_string_hashentry);

        if (not(SvOK(possible_string_hashentry_value)))
        {
            possible_string_hashentry_key = hv_iterkeysv(possible_string_hashentry);
            possible_string_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_string_hashentry_key)));  // escape key string for error message
            croak("\nERROR EPVHVRV02, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring value expected but undefined/null value found at key '%s',\nin variable %s from subroutine %s,\ncroaking", possible_string_hashentry_key_string.c_str(), variable_name, subroutine_name);
        }
        if (not(SvPOKp(possible_string_hashentry_value)))
        {
            possible_string_hashentry_key = hv_iterkeysv(possible_string_hashentry);
            possible_string_hashentry_key_string = escape_backslash_singlequote(string(SvPV_nolen(possible_string_hashentry_key)));  // escape key string for error message
            croak("\nERROR EPVHVRV03, TYPE-CHECKING MISMATCH, CPPOPS_PERLTYPES & CPPOPS_CPPTYPES:\nstring value expected but non-string value found at key '%s',\nin variable %s from subroutine %s,\ncroaking", possible_string_hashentry_key_string.c_str(), variable_name, subroutine_name);
        }
    }
}

// [[[ TYPEMAP PACK/UNPACK FOR __CPP__TYPES ]]]
// [[[ TYPEMAP PACK/UNPACK FOR __CPP__TYPES ]]]
// [[[ TYPEMAP PACK/UNPACK FOR __CPP__TYPES ]]]

# ifdef __CPP__TYPES

// convert from (Perl SV containing reference to (Perl HV of (Perl SVs containing IVs))) to (C++ std::unordered_map of integers)
integer_hashref XS_unpack_integer_hashref(SV* input_hvref)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_integer_hashref(), top of subroutine\n");
//  integer_hashref_CHECK(input_hvref);
    integer_hashref_CHECKTRACE(input_hvref, "input_hvref", "XS_unpack_integer_hashref()");

    HV* input_hv;
    integer input_hv_num_keys;
    integer i;
    HE* input_hv_entry;
    SV* input_hv_entry_key;
    SV* input_hv_entry_value;
    integer_hashref output_unordered_map;

    input_hv = (HV*)SvRV(input_hvref);

    input_hv_num_keys = hv_iterinit(input_hv);
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_integer_hashref(), have input_hv_num_keys = %"INTEGER"\n", input_hv_num_keys);

    // UNORDERED MAP ENTRY ASSIGNMENT, OPTION A, SUBSCRIPT, KNOWN SIZE: unordered_map has programmer-provided const size or compiler-guessable size,
    // reserve() ahead of time to avoid resizing and rehashing in for() loop
    output_unordered_map.reserve((size_t)input_hv_num_keys);

    for (i = 0;  i < input_hv_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        // does not utilize i in entry retrieval
        input_hv_entry = hv_iternext(input_hv);
        // DEV NOTE: hash entry type-checking already done as part of integer_hashref_CHECKTRACE()
//      hashentry_CHECK(input_hv_entry);
//      hashentry_CHECKTRACE(input_hv_entry, "input_hv_entry", "XS_unpack_integer_hashref()");

        input_hv_entry_key = hv_iterkeysv(input_hv_entry);
        input_hv_entry_value = hv_iterval(input_hv, input_hv_entry);
        // DEV NOTE: integer type-checking already done as part of integer_hashref_CHECKTRACE()
//      integer_CHECK(input_hv_entry_value);
//      integer_CHECKTRACE(input_hv_entry_value, (char*)((string)"input_hv_entry_value at key '" + (string)SvPV_nolen(input_hv_entry_key) + "'").c_str(), "XS_unpack_integer_hashref()");

        // UNORDERED MAP ENTRY ASSIGNMENT, OPTION A, SUBSCRIPT, KNOWN SIZE: l-value subscript notation with no further reserve(); does not utilize i in assignment
        output_unordered_map[SvPV_nolen(input_hv_entry_key)] = SvIV(input_hv_entry_value);
    }

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_integer_hashref(), after for() loop, have output_unordered_map.size() = %"INTEGER"\n", output_unordered_map.size());
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_integer_hashref(), bottom of subroutine\n");

    return(output_unordered_map);
}

// convert from (C++ std::unordered_map of integers) to (Perl SV containing reference to (Perl HV of (Perl SVs containing IVs)))
void XS_pack_integer_hashref(SV* output_hvref, integer_hashref input_unordered_map)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_integer_hashref(), top of subroutine\n");

    HV* output_hv = newHV();  // initialize output hash to empty
    integer input_unordered_map_num_keys = input_unordered_map.size();
    integer_hashref_const_iterator i;
    SV* temp_sv_pointer;

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_integer_hashref(), have input_unordered_map_num_keys = %"INTEGER"\n", input_unordered_map_num_keys);

    if (input_unordered_map_num_keys > 0)
    {
        for (i = input_unordered_map.begin();  i != input_unordered_map.end();  ++i)
            { hv_store(output_hv, (const char*)((i->first).c_str()), (U32)((i->first).size()), newSViv(i->second), (U32)0); }
    }
//  else warn("in CPPOPS_CPPTYPES XS_pack_integer_hashref(), hash was empty, returning empty hash via newHV()");

    temp_sv_pointer = newSVrv(output_hvref, NULL);    // upgrade output stack SV to an RV
    SvREFCNT_dec(temp_sv_pointer);       // discard temporary pointer
    SvRV(output_hvref) = (SV*)output_hv;       // make output stack RV pointer at our output HV

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_integer_hashref(), bottom of subroutine\n");
}

// convert from (Perl SV containing reference to (Perl HV of (Perl SVs containing NVs))) to (C++ std::unordered_map of doubles)
number_hashref XS_unpack_number_hashref(SV* input_hvref)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_number_hashref(), top of subroutine\n");
//  number_hashref_CHECK(input_hvref);
    number_hashref_CHECKTRACE(input_hvref, "input_hvref", "XS_unpack_number_hashref()");

    HV* input_hv;
    integer input_hv_num_keys;
    integer i;
    HE* input_hv_entry;
    SV* input_hv_entry_key;
    SV* input_hv_entry_value;
    number_hashref output_unordered_map;

    input_hv = (HV*)SvRV(input_hvref);

    input_hv_num_keys = hv_iterinit(input_hv);
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_number_hashref(), have input_hv_num_keys = %"INTEGER"\n", input_hv_num_keys);

    // UNORDERED MAP ENTRY ASSIGNMENT, OPTION A, SUBSCRIPT, KNOWN SIZE: unordered_map has programmer-provided const size or compiler-guessable size,
    // reserve() ahead of time to avoid resizing and rehashing in for() loop
    output_unordered_map.reserve((size_t)input_hv_num_keys);

    for (i = 0;  i < input_hv_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        // does not utilize i in entry retrieval
        input_hv_entry = hv_iternext(input_hv);
        input_hv_entry_key = hv_iterkeysv(input_hv_entry);
        input_hv_entry_value = hv_iterval(input_hv, input_hv_entry);

        // UNORDERED MAP ENTRY ASSIGNMENT, OPTION A, SUBSCRIPT, KNOWN SIZE: l-value subscript notation with no further reserve(); does not utilize i in assignment
        output_unordered_map[SvPV_nolen(input_hv_entry_key)] = SvNV(input_hv_entry_value);
    }

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_number_hashref(), after for() loop, have output_unordered_map.size() = %"INTEGER"\n", output_unordered_map.size());
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_number_hashref(), bottom of subroutine\n");

    return(output_unordered_map);
}

// convert from (C++ std::unordered_map of doubles) to (Perl SV containing reference to (Perl HV of (Perl SVs containing NVs)))
void XS_pack_number_hashref(SV* output_hvref, number_hashref input_unordered_map)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_number_hashref(), top of subroutine\n");

    HV* output_hv = newHV();  // initialize output hash to empty
    integer input_unordered_map_num_keys = input_unordered_map.size();
    number_hashref_const_iterator i;
    SV* temp_sv_pointer;

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_number_hashref(), have input_unordered_map_num_keys = %"INTEGER"\n", input_unordered_map_num_keys);

    if (input_unordered_map_num_keys > 0)
    {
        for (i = input_unordered_map.begin();  i != input_unordered_map.end();  ++i)
            { hv_store(output_hv, (const char*)((i->first).c_str()), (U32)((i->first).size()), newSVnv(i->second), (U32)0); }
    }
//  else warn("in CPPOPS_CPPTYPES XS_pack_number_hashref(), hash was empty, returning empty hash via newHV()");

    temp_sv_pointer = newSVrv(output_hvref, NULL);    // upgrade output stack SV to an RV
    SvREFCNT_dec(temp_sv_pointer);       // discard temporary pointer
    SvRV(output_hvref) = (SV*)output_hv;       // make output stack RV pointer at our output HV

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_number_hashref(), bottom of subroutine\n");
}

// convert from (Perl SV containing reference to (Perl HV of (Perl SVs containing PVs))) to (C++ std::unordered_map of strings)
string_hashref XS_unpack_string_hashref(SV* input_hvref)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_string_hashref(), top of subroutine\n");
//  string_hashref_CHECK(input_hvref);
    string_hashref_CHECKTRACE(input_hvref, "input_hvref", "XS_unpack_string_hashref()");

    HV* input_hv;
    integer input_hv_num_keys;
    integer i;
    HE* input_hv_entry;
    SV* input_hv_entry_key;
    SV* input_hv_entry_value;
    string_hashref output_unordered_map;

    input_hv = (HV*)SvRV(input_hvref);

    input_hv_num_keys = hv_iterinit(input_hv);
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_string_hashref(), have input_hv_num_keys = %"INTEGER"\n", input_hv_num_keys);

    // UNORDERED MAP ENTRY ASSIGNMENT, OPTION A, SUBSCRIPT, KNOWN SIZE: unordered_map has programmer-provided const size or compiler-guessable size,
    // reserve() ahead of time to avoid resizing and rehashing in for() loop
    output_unordered_map.reserve((size_t)input_hv_num_keys);

    for (i = 0;  i < input_hv_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        // does not utilize i in entry retrieval
        input_hv_entry = hv_iternext(input_hv);
        input_hv_entry_key = hv_iterkeysv(input_hv_entry);
        input_hv_entry_value = hv_iterval(input_hv, input_hv_entry);

        // UNORDERED MAP ENTRY ASSIGNMENT, OPTION A, SUBSCRIPT, KNOWN SIZE: l-value subscript notation with no further reserve(); does not utilize i in assignment
        output_unordered_map[SvPV_nolen(input_hv_entry_key)] = SvPV_nolen(input_hv_entry_value);
    }

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_string_hashref(), after for() loop, have output_unordered_map.size() = %"INTEGER"\n", output_unordered_map.size());
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_unpack_string_hashref(), bottom of subroutine\n");

    return(output_unordered_map);
}

// convert from (C++ std::unordered_map of strings) to (Perl SV containing reference to (Perl HV of (Perl SVs containing PVs)))
void XS_pack_string_hashref(SV* output_hvref, string_hashref input_unordered_map)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_string_hashref(), top of subroutine\n");

    HV* output_hv = newHV();  // initialize output hash to empty
    integer input_unordered_map_num_keys = input_unordered_map.size();
    string_hashref_const_iterator i;
    SV* temp_sv_pointer;

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_string_hashref(), have input_unordered_map_num_keys = %"INTEGER"\n", input_unordered_map_num_keys);

    if (input_unordered_map_num_keys > 0)
    {
        for (i = input_unordered_map.begin();  i != input_unordered_map.end();  ++i)
            { hv_store(output_hv, (const char*)((i->first).c_str()), (U32)((i->first).size()), newSVpv((i->second).c_str(), 0), (U32)0); }
    }
//  else warn("in CPPOPS_CPPTYPES XS_pack_string_hashref(), hash was empty, returning empty hash via newHV()");

    temp_sv_pointer = newSVrv(output_hvref, NULL);    // upgrade output stack SV to an RV
    SvREFCNT_dec(temp_sv_pointer);       // discard temporary pointer
    SvRV(output_hvref) = (SV*)output_hv;       // make output stack RV pointer at our output HV

//  fprintf(stderr, "in CPPOPS_CPPTYPES XS_pack_string_hashref(), bottom of subroutine\n");
}

# endif

// [[[ STRINGIFY ]]]
// [[[ STRINGIFY ]]]
// [[[ STRINGIFY ]]]

# ifdef __PERL__TYPES

// DEV NOTE: 1-D format levels are 1 less than 2-D format levels

// call actual stringify routine, format level -2 (compact), indent level 0
SV* integer_hashref_to_string_compact(SV* input_hvref) {
    return integer_hashref_to_string_format(input_hvref, newSViv(-2), newSViv(0));
}

// call actual stringify routine, format level -1 (normal), indent level 0, DEFAULT
SV* integer_hashref_to_string(SV* input_hvref) {
    return integer_hashref_to_string_format(input_hvref, newSViv(-1), newSViv(0));
}

// call actual stringify routine, format level 0 (pretty), indent level 0
SV* integer_hashref_to_string_pretty(SV* input_hvref) {
    return integer_hashref_to_string_format(input_hvref, newSViv(0), newSViv(0));
}

// call actual stringify routine, format level 1 (expand), indent level 0
SV* integer_hashref_to_string_expand(SV* input_hvref) {
    return integer_hashref_to_string_format(input_hvref, newSViv(1), newSViv(0));
}

// convert from (Perl SV containing RV to (Perl HV of (Perl SVs containing IVs))) to Perl-parsable (Perl SV containing PV)
SV* integer_hashref_to_string_format(SV* input_hvref, SV* format_level, SV* indent_level)
{
//  fprintf(stderr, "in CPPOPS_PERLTYPES integer_hashref_to_string(), top of subroutine\n");
//  integer_hashref_CHECK(input_hvref);
    integer_hashref_CHECKTRACE(input_hvref, "input_hvref", "integer_hashref_to_string()");

    // declare local variables
    HV* input_hv;
    integer input_hv_num_keys;
    integer i;
    boolean i_is_0 = 1;
    HE* input_hv_entry;
    SV* input_hv_entry_key;
    string input_hv_entry_key_string;
    SV* input_hv_entry_value;
    SV* output_sv = newSVpv("", 0);

    // generate indent
    SV* indent = newSVpv("", 0);
    for (i = 0; i < SvIV(indent_level); i++) { sv_catpvn(indent, "    ", 4); }

    // compute length of (number of keys in) input hash
    input_hv = (HV*)SvRV(input_hvref);
    input_hv_num_keys = hv_iterinit(input_hv);
//  fprintf(stderr, "in CPPOPS_PERLTYPES integer_hashref_to_string(), have input_hv_num_keys = %"INTEGER"\n", input_hv_num_keys);

    // pre-begin with optional indent, depending on format level
    if (SvIV(format_level) >= 1) { sv_catsv(output_sv, indent); }

    // begin output string with left-curly-brace, as required for all RPerl hashes
    sv_setpvn(output_sv, "{", 1);

    // loop through all hash keys
    for (i = 0;  i < input_hv_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
//      fprintf(stderr, "in CPPOPS_PERLTYPES integer_hashref_to_string(), top of loop i = %"INTEGER"\n", i);

        // does not utilize i in entry retrieval
        input_hv_entry = hv_iternext(input_hv);
        // DEV NOTE: hash entry type-checking already done as part of integer_hashref_CHECKTRACE()
//      hashentry_CHECK(input_hv_entry);
//      hashentry_CHECKTRACE(input_hv_entry, "input_hv_entry", "integer_hashref_to_string()");

        // retrieve input hash's entry value at key
        input_hv_entry_key = hv_iterkeysv(input_hv_entry);
        input_hv_entry_value = hv_iterval(input_hv, input_hv_entry);
        // DEV NOTE: integer type-checking already done as part of integer_hashref_CHECKTRACE()
//      integer_CHECK(input_hv_entry_value);
//      integer_CHECKTRACE(input_hv_entry_value, (char*)((string)"input_hv_entry_value at key '" + (string)SvPV_nolen(input_hv_entry_key) + "'").c_str(), "integer_hashref_to_string()");

        // append comma to output string for all elements except index 0
        if (i_is_0) { i_is_0 = 0; }
        else        { sv_catpvn(output_sv, ",", 1); }

        // append newline-indent-tab or space, depending on format level
        if      (SvIV(format_level) >=  1) { sv_catpvn(output_sv, "\n", 1);  sv_catsv(output_sv, indent);  sv_catpvn(output_sv, "    ", 4); }
        else if (SvIV(format_level) >= -1) { sv_catpvn(output_sv, " ", 1); }

        // escape key string
        input_hv_entry_key_string = escape_backslash_singlequote(string(SvPV_nolen(input_hv_entry_key)));

        // DEV NOTE: emulate Data::Dumper & follow PBP by using single quotes for key strings
//        sv_catpvf(output_sv, "'%s'", SvPV_nolen(input_hv_entry_key));  // alternative form
        sv_catpvf(output_sv, "'%s'", input_hv_entry_key_string.c_str());

        // append spaces before and after fat arrow AKA fat comma, depending on format level
        if      (SvIV(format_level) >= -1) { sv_catpvn(output_sv, " => ", 4); }
        else                               { sv_catpvn(output_sv, "=>", 2); }

//        sv_catpvf(output_sv, "%"INTEGER"", (integer)SvIV(input_hv_entry_value));  // NO UNDERSCORES
        sv_catsv(output_sv, integer_to_string(input_hv_entry_value));  // YES UNDERSCORES
    }

    // append newline-indent or space, depending on format level
    if      (SvIV(format_level) >=  1) { sv_catpvn(output_sv, "\n", 1);  sv_catsv(output_sv, indent); }
    else if (SvIV(format_level) >= -1) { sv_catpvn(output_sv, " ", 1); }

    // end output string with right-curly-brace, as required for all RPerl hashes
    sv_catpvn(output_sv, "}", 1);

//  fprintf(stderr, "in CPPOPS_PERLTYPES integer_hashref_to_string(), after for() loop, have output_sv =\n%s\n", SvPV_nolen(output_sv));
//  fprintf(stderr, "in CPPOPS_PERLTYPES integer_hashref_to_string(), bottom of subroutine\n");

    return(output_sv);
}

// DEV NOTE: 1-D format levels are 1 less than 2-D format levels

// call actual stringify routine, format level -2 (compact), indent level 0
SV* number_hashref_to_string_compact(SV* input_hvref) {
    return number_hashref_to_string_format(input_hvref, newSViv(-2), newSViv(0));
}

// call actual stringify routine, format level -1 (normal), indent level 0, DEFAULT
SV* number_hashref_to_string(SV* input_hvref) {
    return number_hashref_to_string_format(input_hvref, newSViv(-1), newSViv(0));
}

// call actual stringify routine, format level 0 (pretty), indent level 0
SV* number_hashref_to_string_pretty(SV* input_hvref) {
    return number_hashref_to_string_format(input_hvref, newSViv(0), newSViv(0));
}

// call actual stringify routine, format level 1 (expand), indent level 0
SV* number_hashref_to_string_expand(SV* input_hvref) {
    return number_hashref_to_string_format(input_hvref, newSViv(1), newSViv(0));
}

// convert from (Perl SV containing RV to (Perl HV of (Perl SVs containing NVs))) to Perl-parsable (Perl SV containing PV)
SV* number_hashref_to_string_format(SV* input_hvref, SV* format_level, SV* indent_level)
{
//  fprintf(stderr, "in CPPOPS_PERLTYPES number_hashref_to_string(), top of subroutine\n");
//  number_hashref_CHECK(input_hvref);
    number_hashref_CHECKTRACE(input_hvref, "input_hvref", "number_hashref_to_string()");

    // declare local variables
    HV* input_hv;
    integer input_hv_num_keys;
    integer i;
    boolean i_is_0 = 1;
    HE* input_hv_entry;
    SV* input_hv_entry_key;
    string input_hv_entry_key_string;
    SV* input_hv_entry_value;
    SV* output_sv = newSV(0);

    // NEED ANSWER: do we actually need to be using ostringstream here for precision, since the actual numbers are being stringified by number_to_string() below???
    ostringstream temp_stream;
    temp_stream.precision(std::numeric_limits<double>::digits10);

    // generate indent
    SV* indent = newSVpv("", 0);
    for (i = 0; i < SvIV(indent_level); i++) { sv_catpvn(indent, "    ", 4); }

    // compute length of (number of keys in) input hash
    input_hv = (HV*)SvRV(input_hvref);
    input_hv_num_keys = hv_iterinit(input_hv);
//  fprintf(stderr, "in CPPOPS_PERLTYPES number_hashref_to_string(), have input_hv_num_keys = %"INTEGER"\n", input_hv_num_keys);

    // pre-begin with optional indent, depending on format level
    if (SvIV(format_level) >= 1) { temp_stream << SvPV_nolen(indent); }

    // begin output string with left-curly-brace, as required for all RPerl hashes
    temp_stream << "{";

    // loop through all hash keys
    for (i = 0;  i < input_hv_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        // does not utilize i in entry retrieval
        input_hv_entry = hv_iternext(input_hv);
        input_hv_entry_key = hv_iterkeysv(input_hv_entry);
        input_hv_entry_value = hv_iterval(input_hv, input_hv_entry);

        // append comma to output string for all elements except index 0
        if (i_is_0) { i_is_0 = 0; }
        else        { temp_stream << ","; }

        // append newline-indent-tab or space, depending on format level
        if      (SvIV(format_level) >=  1) { temp_stream << "\n" << SvPV_nolen(indent) << "    "; }
        else if (SvIV(format_level) >= -1) { temp_stream << " "; }

        // escape key string
        input_hv_entry_key_string = escape_backslash_singlequote(string(SvPV_nolen(input_hv_entry_key)));

        // DEV NOTE: emulate Data::Dumper & follow PBP by using single quotes for key strings
//        temp_stream << "'" << SvPV_nolen(input_hv_entry_key) << "'";  // alternative form
        temp_stream << "'" << input_hv_entry_key_string.c_str() << "'";

        // append spaces before and after fat arrow AKA fat comma, depending on format level
        if      (SvIV(format_level) >= -1) { temp_stream << " => "; }
        else                               { temp_stream << "=>"; }

//      temp_stream << (double)SvNV(input_hv_entry_value);  // NO UNDERSCORES
        temp_stream << (string)SvPV_nolen(number_to_string(input_hv_entry_value));  // YES UNDERSCORES
    }

    // append newline-indent or space, depending on format level
    if      (SvIV(format_level) >=  1) { temp_stream << "\n" << SvPV_nolen(indent); }
    else if (SvIV(format_level) >= -1) { temp_stream << " "; }

    // end output string with right-curly-brace, as required for all RPerl hashes
    temp_stream << "}";
    sv_setpv(output_sv, (char*)(temp_stream.str().c_str()));

//  fprintf(stderr, "in CPPOPS_PERLTYPES number_hashref_to_string(), after for() loop, have output_sv =\n%s\n", SvPV_nolen(output_sv));
//  fprintf(stderr, "in CPPOPS_PERLTYPES number_hashref_to_string(), bottom of subroutine\n");

    return(output_sv);
}

// DEV NOTE: 1-D format levels are 1 less than 2-D format levels

// call actual stringify routine, format level -2 (compact), indent level 0
SV* string_hashref_to_string_compact(SV* input_hvref) {
    return string_hashref_to_string_format(input_hvref, newSViv(-2), newSViv(0));
}

// call actual stringify routine, format level -1 (normal), indent level 0, DEFAULT
SV* string_hashref_to_string(SV* input_hvref) {
    return string_hashref_to_string_format(input_hvref, newSViv(-1), newSViv(0));
}

// call actual stringify routine, format level 0 (pretty), indent level 0
SV* string_hashref_to_string_pretty(SV* input_hvref) {
    return string_hashref_to_string_format(input_hvref, newSViv(0), newSViv(0));
}

// call actual stringify routine, format level 1 (expand), indent level 0
SV* string_hashref_to_string_expand(SV* input_hvref) {
    return string_hashref_to_string_format(input_hvref, newSViv(1), newSViv(0));
}

// convert from (Perl SV containing RV to (Perl HV of (Perl SVs containing PVs))) to Perl-parsable (Perl SV containing PV)
SV* string_hashref_to_string_format(SV* input_hvref, SV* format_level, SV* indent_level)
{
//  fprintf(stderr, "in CPPOPS_PERLTYPES string_hashref_to_string(), top of subroutine\n");
//  string_hashref_CHECK(input_hvref);
    string_hashref_CHECKTRACE(input_hvref, "input_hvref", "string_hashref_to_string()");

    // declare local variables
    HV* input_hv;
    integer input_hv_num_keys;
    integer i;
    boolean i_is_0 = 1;
    HE* input_hv_entry;
    SV* input_hv_entry_key;
    string input_hv_entry_key_string;
    SV* input_hv_entry_value;
    string input_hv_entry_value_string;
    SV* output_sv = newSVpv("", 0);

    // generate indent
    SV* indent = newSVpv("", 0);
    for (i = 0; i < SvIV(indent_level); i++) { sv_catpvn(indent, "    ", 4); }

    // compute length of (number of keys in) input hash
    input_hv = (HV*)SvRV(input_hvref);
    input_hv_num_keys = hv_iterinit(input_hv);
//  fprintf(stderr, "in CPPOPS_PERLTYPES string_hashref_to_string(), have input_hv_num_keys = %"INTEGER"\n", input_hv_num_keys);

    // pre-begin with optional indent, depending on format level
    if (SvIV(format_level) >= 1) { sv_catsv(output_sv, indent); }

    // begin output string with left-curly-brace, as required for all RPerl hashes
    sv_setpvn(output_sv, "{", 1);

    // loop through all hash keys
    for (i = 0;  i < input_hv_num_keys;  ++i)  // incrementing iteration, iterator i not actually used in loop body
    {
        // does not utilize i in entry retrieval
        input_hv_entry = hv_iternext(input_hv);
        input_hv_entry_key = hv_iterkeysv(input_hv_entry);
        input_hv_entry_value = hv_iterval(input_hv, input_hv_entry);

        // append comma to output string for all elements except index 0
        if (i_is_0) { i_is_0 = 0; }
        else        { sv_catpvn(output_sv, ",", 1); }

        // append newline-indent-tab or space, depending on format level
        if      (SvIV(format_level) >=  1) { sv_catpvn(output_sv, "\n", 1);  sv_catsv(output_sv, indent);  sv_catpvn(output_sv, "    ", 4); }
        else if (SvIV(format_level) >= -1) { sv_catpvn(output_sv, " ", 1); }

        // escape key string
        input_hv_entry_key_string = escape_backslash_singlequote(string(SvPV_nolen(input_hv_entry_key)));

        // escape value string
        input_hv_entry_value_string = escape_backslash_singlequote(string(SvPV_nolen(input_hv_entry_value)));

        // DEV NOTE: emulate Data::Dumper & follow PBP by using single quotes for key strings
//        sv_catpvf(output_sv, "'%s'", SvPV_nolen(input_hv_entry_key));  // alternative form
        sv_catpvf(output_sv, "'%s'", input_hv_entry_key_string.c_str());

        // append spaces before and after fat arrow AKA fat comma, depending on format level
        if      (SvIV(format_level) >= -1) { sv_catpvn(output_sv, " => ", 4); }
        else                               { sv_catpvn(output_sv, "=>", 2); }

//      sv_catpvf(output_sv, "'%s'", SvPV_nolen(input_hv_entry_value));  // alternative form
        sv_catpvf(output_sv, "'%s'", input_hv_entry_value_string.c_str());
    }

    // append newline-indent or space, depending on format level
    if      (SvIV(format_level) >=  1) { sv_catpvn(output_sv, "\n", 1);  sv_catsv(output_sv, indent); }
    else if (SvIV(format_level) >= -1) { sv_catpvn(output_sv, " ", 1); }

    // end output string with right-curly-brace, as required for all RPerl hashes
    sv_catpvn(output_sv, "}", 1);

//  fprintf(stderr, "in CPPOPS_PERLTYPES string_hashref_to_string(), after for() loop, have output_sv =\n%s\n", SvPV_nolen(output_sv));
//  fprintf(stderr, "in CPPOPS_PERLTYPES string_hashref_to_string(), bottom of subroutine\n");

    return(output_sv);
}

# elif defined __CPP__TYPES

// DEV NOTE: 1-D format levels are 1 less than 2-D format levels

// call actual stringify routine, format level -2 (compact), indent level 0
string integer_hashref_to_string_compact(integer_hashref input_unordered_map)
{
    return integer_hashref_to_string_format(input_unordered_map, -2, 0);
}

// call actual stringify routine, format level -1 (normal), indent level 0, DEFAULT
string integer_hashref_to_string(integer_hashref input_unordered_map)
{
    return integer_hashref_to_string_format(input_unordered_map, -1, 0);
}

// call actual stringify routine, format level 0 (pretty), indent level 0
string integer_hashref_to_string_pretty(integer_hashref input_unordered_map)
{
    return integer_hashref_to_string_format(input_unordered_map, 0, 0);
}

// call actual stringify routine, format level 1 (expand), indent level 0
string integer_hashref_to_string_expand(integer_hashref input_unordered_map)
{
    return integer_hashref_to_string_format(input_unordered_map, 1, 0);
}

// convert from (C++ std::unordered_map of integers) to Perl-parsable (C++ std::string)
string integer_hashref_to_string_format(integer_hashref input_unordered_map, integer format_level, integer indent_level)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES integer_hashref_to_string(), top of subroutine\n");

    // declare local variables
    ostringstream output_stream;
    integer_hashref_const_iterator i;
    boolean i_is_0 = 1;
    string key_string;

    // generate indent
    string indent = "";
    for (integer indent_i = 0; indent_i < indent_level; indent_i++) { indent += "    "; }

    // pre-begin with optional indent, depending on format level
    if (format_level >= 1) { output_stream << indent; }

    // begin output string with left-curly-brace, as required for all RPerl hashes
    output_stream << '{';

    // loop through all hash keys
    for (i = input_unordered_map.begin();  i != input_unordered_map.end();  ++i)
    {
        // append comma to output string for all elements except index 0
        if (i_is_0) { i_is_0 = 0; }
        else        { output_stream << ','; }

        // append newline-indent-tab or space, depending on format level
        if      (format_level >=  1) { output_stream << endl << indent << "    "; }
        else if (format_level >= -1) { output_stream << ' '; }

        // escape key string
        key_string = escape_backslash_singlequote(i->first);

        // DEV NOTE: emulate Data::Dumper & follow PBP by using single quotes for key strings
//        output_stream << "'" << (i->first).c_str() << "'";  // alternative format
        output_stream << "'" << key_string.c_str() << "'";

        // append spaces before and after fat arrow AKA fat comma, depending on format level
        if (format_level >= -1) { output_stream << " => "; }
        else                    { output_stream << "=>"; }

//        output_stream << i->second;  // NO UNDERSCORES
        output_stream << integer_to_string(i->second);  // YES UNDERSCORES
    }

    // append newline-indent or space, depending on format level
    if      (format_level >=  1) { output_stream << endl << indent; }
    else if (format_level >= -1) { output_stream << ' '; }

    // end output string with right-curly-brace, as required for all RPerl hashes
    output_stream << '}';

//  fprintf(stderr, "in CPPOPS_CPPTYPES integer_hashref_to_string(), after for() loop, have output_stream =\n%s\n", (char*)(output_stream.str().c_str()));
//  fprintf(stderr, "in CPPOPS_CPPTYPES integer_hashref_to_string(), bottom of subroutine\n");

    return(output_stream.str());
}

// DEV NOTE: 1-D format levels are 1 less than 2-D format levels

// call actual stringify routine, format level -2 (compact), indent level 0
string number_hashref_to_string_compact(number_hashref input_unordered_map)
{
    return number_hashref_to_string_format(input_unordered_map, -2, 0);
}

// call actual stringify routine, format level -1 (normal), indent level 0, DEFAULT
string number_hashref_to_string(number_hashref input_unordered_map)
{
    return number_hashref_to_string_format(input_unordered_map, -1, 0);
}

// call actual stringify routine, format level 0 (pretty), indent level 0
string number_hashref_to_string_pretty(number_hashref input_unordered_map)
{
    return number_hashref_to_string_format(input_unordered_map, 0, 0);
}

// call actual stringify routine, format level 1 (expand), indent level 0
string number_hashref_to_string_expand(number_hashref input_unordered_map)
{
    return number_hashref_to_string_format(input_unordered_map, 1, 0);
}

// convert from (C++ std::unordered_map of doubles) to Perl-parsable (C++ std::string)
string number_hashref_to_string_format(number_hashref input_unordered_map, integer format_level, integer indent_level)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES number_hashref_to_string(), top of subroutine\n");

    // declare local variables
    ostringstream output_stream;
    number_hashref_const_iterator i;
    boolean i_is_0 = 1;
    string key_string;

    // NEED ANSWER: do we actually need to be using ostringstream here for precision, since the actual numbers are being stringified by number_to_string() below???
    output_stream.precision(std::numeric_limits<double>::digits10);

    // generate indent
    string indent = "";
    for (integer indent_i = 0; indent_i < indent_level; indent_i++) { indent += "    "; }

    // pre-begin with optional indent, depending on format level
    if (format_level >= 1) { output_stream << indent; }

    // begin output string with left-curly-brace, as required for all RPerl hashes
    output_stream << '{';

    // loop through all hash keys
    for (i = input_unordered_map.begin();  i != input_unordered_map.end();  ++i)
    {
        // append comma to output string for all elements except index 0
        if (i_is_0) { i_is_0 = 0; }
        else        { output_stream << ','; }

        // append newline-indent-tab or space, depending on format level
        if      (format_level >=  1) { output_stream << endl << indent << "    "; }
        else if (format_level >= -1) { output_stream << ' '; }

        // escape key string
        key_string = escape_backslash_singlequote(i->first);

        // DEV NOTE: emulate Data::Dumper & follow PBP by using single quotes for key strings
//        output_stream << "'" << (i->first).c_str() << "'";  // alternative format
        output_stream << "'" << key_string.c_str() << "'";

        // append spaces before and after fat arrow AKA fat comma, depending on format level
        if (format_level >= -1) { output_stream << " => "; }
        else                    { output_stream << "=>"; }

//        output_stream << i->second;  // NO UNDERSCORES
        output_stream << number_to_string(i->second);  // YES UNDERSCORES
    }

    // append newline-indent or space, depending on format level
    if      (format_level >=  1) { output_stream << endl << indent; }
    else if (format_level >= -1) { output_stream << ' '; }

    // end output string with right-curly-brace, as required for all RPerl hashes
    output_stream << '}';

//  fprintf(stderr, "in CPPOPS_CPPTYPES number_hashref_to_string(), after for() loop, have output_stream =\n%s\n", (char*)(output_stream.str().c_str()));
//  fprintf(stderr, "in CPPOPS_CPPTYPES number_hashref_to_string(), bottom of subroutine\n");

    return(output_stream.str());
}

// DEV NOTE: 1-D format levels are 1 less than 2-D format levels

// call actual stringify routine, format level -2 (compact), indent level 0
string string_hashref_to_string_compact(string_hashref input_unordered_map)
{
    return string_hashref_to_string_format(input_unordered_map, -2, 0);
}

// call actual stringify routine, format level -1 (normal), indent level 0, DEFAULT
string string_hashref_to_string(string_hashref input_unordered_map)
{
    return string_hashref_to_string_format(input_unordered_map, -1, 0);
}

// call actual stringify routine, format level 0 (pretty), indent level 0
string string_hashref_to_string_pretty(string_hashref input_unordered_map)
{
    return string_hashref_to_string_format(input_unordered_map, 0, 0);
}

// call actual stringify routine, format level 1 (expand), indent level 0
string string_hashref_to_string_expand(string_hashref input_unordered_map)
{
    return string_hashref_to_string_format(input_unordered_map, 1, 0);
}

// convert from (C++ std::unordered_map of std::strings) to Perl-parsable (C++ std::string)
string string_hashref_to_string_format(string_hashref input_unordered_map, integer format_level, integer indent_level)
{
//  fprintf(stderr, "in CPPOPS_CPPTYPES string_hashref_to_string(), top of subroutine\n");

    // declare local variables
    string output_string;
    string_hashref_const_iterator i;
    boolean i_is_0 = 1;
    string key_string;
    string value_string;

    // generate indent
    string indent = "";
    for (integer indent_i = 0; indent_i < indent_level; indent_i++) { indent += "    "; }

    // pre-begin with optional indent, depending on format level
    if (format_level >= 1) { output_string += indent; }

    // begin output string with left-curly-brace, as required for all RPerl hashes
    output_string = "{";

    // loop through all hash keys
    for (i = input_unordered_map.begin();  i != input_unordered_map.end();  ++i)
    {
        // append comma to output string for all elements except index 0
        if (i_is_0) { i_is_0 = 0; }
        else        { output_string += ","; }

        // append newline-indent-tab or space, depending on format level
        if      (format_level >=  1) { output_string += "\n" + indent + "    "; }
        else if (format_level >= -1) { output_string += " "; }

        // escape key string
        key_string = escape_backslash_singlequote(i->first);

        // escape value string
        value_string = escape_backslash_singlequote(i->second);

        // DEV NOTE: emulate Data::Dumper & follow PBP by using single quotes for key strings
//      output_string += "'" + (string)(i->first).c_str() + "'";  // alternative format
        output_string += "'" + key_string + "'";

        // append spaces before and after fat arrow AKA fat comma, depending on format level
        if (format_level >= -1) { output_string += " => "; }
        else                    { output_string += "=>"; }

//      output_string += "'" + (string)(i->second) + "'";  // alternative format
        output_string += "'" + value_string + "'";
    }

    // append newline-indent or space, depending on format level
    if      (format_level >=  1) { output_string += "\n" + indent; }
    else if (format_level >= -1) { output_string += " "; }

    // end output string with right-curly-brace, as required for all RPerl hashes
    output_string += "}";

//  fprintf(stderr, "in CPPOPS_CPPTYPES string_hashref_to_string(), after for() loop, have output_string =\n%s\n", output_string.c_str());
//  fprintf(stderr, "in CPPOPS_CPPTYPES string_hashref_to_string(), bottom of subroutine\n");

    return(output_string);
}

# else

Purposefully_die_from_a_compile-time_error,_due_to_neither___PERL__TYPES_nor___CPP__TYPES_being_defined.__We_need_to_define_exactly_one!

# endif

// [[[ TYPE TESTING ]]]
// [[[ TYPE TESTING ]]]
// [[[ TYPE TESTING ]]]

# ifdef __PERL__TYPES

SV* integer_hashref_typetest0(SV* lucky_integers)
{
//  integer_hashref_CHECK(lucky_integers);
    integer_hashref_CHECKTRACE(lucky_integers, "lucky_integers", "integer_hashref_typetest0()");

/*
    HV* lucky_integers_deref = (HV*)SvRV(lucky_integers);
    integer how_lucky = hv_iterinit(lucky_integers_deref);
    integer i;

    for (i = 0;  i < how_lucky;  ++i)
    {
        HE* lucky_integer_entry = hv_iternext(lucky_integers_deref);
        // DEV NOTE: hash entry type-checking already done as part of integer_hashref_CHECKTRACE()
//      hashentry_CHECK(lucky_integer_entry);
//      hashentry_CHECKTRACE(lucky_integer_entry, "lucky_integer_entry", "integer_hashref_typetest0()");

        // DEV NOTE: not using lucky_number variable as in Hash.pm
        // DEV NOTE: integer type-checking already done as part of integer_hashref_CHECKTRACE()
//      integer_CHECK(hv_iterval(lucky_integers_deref, lucky_integer_entry));
//      integer_CHECKTRACE(hv_iterval(lucky_integers_deref, lucky_integer_entry), (char*)((string)"hv_iterval(lucky_integers_deref, lucky_integer_entry) at key '" + (string)SvPV_nolen(hv_iterkeysv(lucky_integer_entry)) + "'").c_str(), "integer_hashref_typetest0()");

//      fprintf(stderr, "in CPPOPS_PERLTYPES integer_hashref_typetest0(), have lucky integer '%s' => %"INTEGER", BARSTOOL\n", SvPV_nolen(hv_iterkeysv(lucky_integer_entry)), (integer)SvIV(hv_iterval(lucky_integers_deref, lucky_integer_entry)));
    }
*/

    return(newSVpvf("%s%s", SvPV_nolen(integer_hashref_to_string(lucky_integers)), "CPPOPS_PERLTYPES"));
}

SV* integer_hashref_typetest1(SV* my_size)
{
//  integer_CHECK(my_size);
    integer_CHECKTRACE(my_size, "my_size", "integer_hashref_typetest1()");
    HV* output_hv = newHV();
    integer i;
    char temp_key[30];

    for (i = 0;  i < SvIV(my_size);  ++i)
    {
        sprintf(temp_key, "CPPOPS_PERLTYPES_funkey%"INTEGER"", i);
        hv_store(output_hv, (const char*)temp_key, (U32)strlen(temp_key), newSViv(i * 5), (U32)0);
//      fprintf(stderr, "in CPPOPS_PERLTYPES integer_hashref_typetest1(), setting entry '%s' => %"INTEGER", BARBAT\n", temp_key, (integer)SvIV(*hv_fetch(output_hv, (const char*)temp_key, (U32)strlen(temp_key), (I32)0)));
    }

    return(newRV_noinc((SV*) output_hv));
}

SV* number_hashref_typetest0(SV* lucky_numbers)
{
//  number_hashref_CHECK(lucky_numbers);
    number_hashref_CHECKTRACE(lucky_numbers, "lucky_numbers", "number_hashref_typetest0()");

/*
    HV* lucky_numbers_deref = (HV*)SvRV(lucky_numbers);
    integer how_lucky = hv_iterinit(lucky_numbers_deref);
    integer i;

    for (i = 0;  i < how_lucky;  ++i)
    {
        HE* lucky_number_entry = hv_iternext(lucky_numbers_deref);
//      fprintf(stderr, "in CPPOPS_PERLTYPES number_hashref_typetest0(), have lucky number '%s' => %"NUMBER", BARSTOOP\n", SvPV_nolen(hv_iterkeysv(lucky_number_entry)), (number)SvNV(hv_iterval(lucky_numbers_deref, lucky_number_entry)));
    }
*/

    return(newSVpvf("%s%s", SvPV_nolen(number_hashref_to_string(lucky_numbers)), "CPPOPS_PERLTYPES"));
}

SV* number_hashref_typetest1(SV* my_size)
{
//  integer_CHECK(my_size);
    integer_CHECKTRACE(my_size, "my_size", "number_hashref_typetest1()");
    HV* output_hv = newHV();
    integer i;
    char temp_key[30];

    for (i = 0;  i < SvIV(my_size);  ++i)
    {
        sprintf(temp_key, "CPPOPS_PERLTYPES_funkey%"INTEGER"", i);
        hv_store(output_hv, (const char*)temp_key, (U32)strlen(temp_key), newSVnv(i * 5.123456789), (U32)0);
//      fprintf(stderr, "in CPPOPS_PERLTYPES number_hashref_typetest1(), setting entry '%s' => %"NUMBER", BARTAB\n", temp_key, (number)SvNV(*hv_fetch(output_hv, (const char*)temp_key, (U32)strlen(temp_key), (I32)0)));
    }

    return(newRV_noinc((SV*) output_hv));
}

SV* string_hashref_typetest0(SV* people)
{
//  string_hashref_CHECK(people);
    string_hashref_CHECKTRACE(people, "people", "string_hashref_typetest0()");

/*
    HV* people_deref = (HV*)SvRV(people);
    integer how_crowded = hv_iterinit(people_deref);
    integer i;

    for (i = 0;  i < how_crowded;  ++i)
    {
        HE* person_entry = hv_iternext(people_deref);
//      fprintf(stderr, "in CPPOPS_PERLTYPES string_hashref_typetest0(), have person '%s' => '%s', BARSPOON\n", (char*)SvPV_nolen(hv_iterkeysv(person_entry)), (char*)SvPV_nolen(hv_iterval(people_deref, person_entry)));
    }
*/

    return(newSVpvf("%s%s", SvPV_nolen(string_hashref_to_string(people)), "CPPOPS_PERLTYPES"));
}

SV* string_hashref_typetest1(SV* my_size)
{
//  integer_CHECK(my_size);
    integer_CHECKTRACE(my_size, "my_size", "string_hashref_typetest1()");
    HV* people = newHV();
    integer i;
    char temp_key[30];

    for (i = 0;  i < SvIV(my_size);  ++i)
    {
        sprintf(temp_key, "CPPOPS_PERLTYPES_Luker_key%"INTEGER"", i);
        hv_store(people, (const char*)temp_key, (U32)strlen(temp_key), newSVpvf("Jeffy Ten! %"INTEGER"/%"INTEGER"", i, (integer)(SvIV(my_size) - 1)), (U32)0);
//      fprintf(stderr, "in CPPOPS_PERLTYPES string_hashref_typetest1(), have temp_key = '%s', just set another Jeffy, BARTAT\n", temp_key);
    }

    return(newRV_noinc((SV*) people));
}

# elif defined __CPP__TYPES

string integer_hashref_typetest0(integer_hashref lucky_integers)
{
    /*
    integer_hashref_const_iterator i;
    for (i = lucky_integers.begin();  i != lucky_integers.end();  ++i)
    {
        fprintf(stderr, "in CPPOPS_CPPTYPES integer_hashref_typetest0(), have lucky integer '%s' => %"INTEGER", BARSTOOL\n", (i->first).c_str(), i->second);
    }
    */
    return(integer_hashref_to_string(lucky_integers) + "CPPOPS_CPPTYPES");
}

integer_hashref integer_hashref_typetest1(integer my_size)
{
    integer_hashref new_unordered_map(my_size);
    integer i;
    string temp_key;
    for (i = 0;  i < my_size;  ++i)
    {
        temp_key = "CPPOPS_CPPTYPES_funkey" + std::to_string(i);
        new_unordered_map[temp_key] = i * 5;
//      fprintf(stderr, "in CPPOPS_CPPTYPES integer_hashref_typetest1(), setting entry '%s' => %"INTEGER", BARSTOOL\n", temp_key.c_str(), new_unordered_map[temp_key]);
    }
    return(new_unordered_map);
}

string number_hashref_typetest0(number_hashref lucky_numbers)
{
    /*
    number_hashref_const_iterator i;
    for (i = lucky_numbers.begin();  i != lucky_numbers.end();  ++i)
    {
        fprintf(stderr, "in CPPOPS_CPPTYPES number_hashref_typetest0(), have lucky number '%s' => %"NUMBER", BARSTOOL\n", (i->first).c_str(), i->second);
    }
    */
    return(number_hashref_to_string(lucky_numbers) + "CPPOPS_CPPTYPES");
}

number_hashref number_hashref_typetest1(integer my_size)
{
    number_hashref new_unordered_map(my_size);
    integer i;
    string temp_key;
    for (i = 0;  i < my_size;  ++i)
    {
        temp_key = "CPPOPS_CPPTYPES_funkey" + std::to_string(i);
        new_unordered_map[temp_key] = i * 5.123456789;
//      fprintf(stderr, "in CPPOPS_CPPTYPES number_hashref_typetest1(), setting entry '%s' => %"NUMBER", BARSTOOL\n", temp_key.c_str(), new_unordered_map[temp_key]);
    }
    return(new_unordered_map);
}

string string_hashref_typetest0(string_hashref people)
{
    /*
    string_hashref_const_iterator i;
    for (i = people.begin();  i != people.end();  ++i)
    {
        fprintf(stderr, "in CPPOPS_CPPTYPES string_hashref_typetest0(), have person '%s' => '%s', STARBOOL\n", (i->first).c_str(), (i->second).c_str());
    }
    */
    return(string_hashref_to_string(people) + "CPPOPS_CPPTYPES");
}

string_hashref string_hashref_typetest1(integer my_size)
{
    string_hashref people;
    integer i;
    people.reserve((size_t)my_size);
    for (i = 0;  i < my_size;  ++i)
    {
        people["CPPOPS_CPPTYPES_Luker_key" + std::to_string(i)] = "Jeffy Ten! " + std::to_string(i) + "/" + std::to_string(my_size - 1);
//      fprintf(stderr, "in CPPOPS_CPPTYPES string_hashref_typetest1(), bottom of for() loop, have i = %"INTEGER", just set another Jeffy!\n", i);
    }
    return(people);
}

# endif

#endif
